from zenlog import log
from dataclasses import dataclass, field
from mccode_antlr.instr import Instr
from mccode_antlr.common import Expr
from nexusformat.nexus import NXfield, NXgroup, NXcollection
from typing import Union, Any


@dataclass
class NXInstr:
    instr: Instr
    declared: dict[str, Expr] = field(default_factory=dict)
    nxlog_root: str = field(default_factory=str)

    def __post_init__(self):
        """Start the C translation to ensure McCode-oddities are handled before any C-code parsing."""
        from mccode_antlr.common import ShapeType, DataType, Value
        from mccode_antlr import Flavor
        from mccode_antlr.translators.c import CTargetVisitor
        from mccode_antlr.translators.c_listener import CDeclarator
        from mccode_antlr.translators.c_listener import evaluate_c_defined_expressions
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False, output=None)
        translator = CTargetVisitor(self.instr, flavor=Flavor.MCSTAS, config=config)
        # translator.instrument_uservars is a list of `CDeclaration` objects, which are named tuples with
        # fields: name type init is_pointer is_array orig
        # translator.component_uservars is a dictionary of lists for each component type of `CDeclaration` objects.

        # only worry about instrument level variables for the moment, and convert the CDeclarations into Expr objects
        def c_declaration_to_expr(dec: CDeclarator) -> Expr:
            expr = Expr(Value(None)) if dec.init is None else Expr.parse(dec.init)
            expr.data_type = DataType.from_name(dec.dtype)
            if dec.is_pointer or dec.is_array:
                expr.shape_type = ShapeType.vector
            return expr

        variables = {dec.name: c_declaration_to_expr(dec) for dec in translator.instrument_uservars}

        # defined as
        # TODO this does not work because the simple "C"-style expression parser doesn't know about pointers
        # Hopefully any %include style lines have been removed at this point.
        all_inits = '\n'.join(init.source for init in self.instr.initialize)
        try:
            variables = evaluate_c_defined_expressions(variables, all_inits)
        except AttributeError:
            log.warn(f'Evaluating INITIALIZE %{{\n{all_inits}%}}\n failed; see preceding errors for hints why. '
                     'This is not an error condition (for now). Continuing')

        self.declared = variables

    def to_nx(self):
        # quick and very dirty:
        return NXfield(str(self.instr))

    def expr2nx(self, expr: Union[str, Expr, Any]):
        from moreniius.utils import link_specifier
        if not isinstance(expr, str) and hasattr(expr, '__iter__'):
            parts = [self.expr2nx(x) for x in expr]
            return tuple(parts) if isinstance(expr, tuple) else parts
        if not isinstance(expr, Expr):
            return expr

        if expr.is_constant:
            return expr.value

        evaluated = expr.evaluate(self.declared)
        if evaluated.is_constant:
            return evaluated.value

        dependencies = [par.name for par in self.instr.parameters if evaluated.depends_on(par.name)]
        if len(dependencies):
            log.warn(f'The expression {expr} depends on instrument parameter(s) {dependencies}\n'
                     f'A link will be inserted for each; make sure their values are stored at {self.nxlog_root}/')
            links = {par: link_specifier(par, f'{self.nxlog_root}/{par}') for par in dependencies}
            return NXcollection(expression=str(expr), **links)

        return str(expr)

    def make_nx(self, nx_class, *args, **kwargs):
        nx_args = [self.expr2nx(expr) for expr in args]
        nx_kwargs = {name: self.expr2nx(expr) for name, expr in kwargs.items()}
        # logged parameters are sometimes requested as NXfield objects, but should be links to the real NXlog
        if nx_class == NXfield and len(nx_args) == 1 and isinstance(nx_args[0], NXcollection) and \
                'expression' in nx_args[0]:
            not_expr = [x for x in nx_args[0] if x != 'expression']
            if len(not_expr) == 1:
                not_expr_arg = nx_args[0][not_expr[0]]
                # if isinstance(not_expr_arg, NXfield):
                #     # We have and want an NXfield, but it might be missing attributes specified in the nx_kwargs
                #     # Passing the keywords to the NXfield constructor versus this method is not identical,
                #     # since some keyword arguments are reserved (and only some of which are noted)
                #     #   Explicit keywords, used in the constructor:
                #     #       value, name, shape, dtype, group, attrs
                #     #   Keywords extracted from the kwargs dict, if present (and all controlling HDF5 file attributes?):
                #     #       chunks, compression, compression_opts, fillvalue, fletcher32, maxshape, scaleoffset, shuffle
                #     # For now, just assume all keywords provided here are _actually_ attributes for the NXfield
                #     # which is an extension of a dict, but can *not* use the update method, since the __setitem__
                #     # method is overridden to wrap inputs in NXattr objects :/
                #     for k, v in nx_kwargs.items():
                #         not_expr_arg.attrs[k] = v
                #     return not_expr_arg

                # TODO make this return an nx_class once we're sure that nx_kwargs is parseable (no mccode_antlr.Expr)
                if all(x in not_expr_arg for x in ('module', 'config')):
                    # This is a file-writer stream directive? So make a group
                    return NXgroup(entries={not_expr[0]: not_expr_arg}, **nx_kwargs)
                print('!!')
                print(not_expr_arg)
                return nx_class(not_expr_arg, **nx_kwargs)
            else:
                raise RuntimeError('Not sure what I should do here')
        return nx_class(*nx_args, **nx_kwargs)




def make_motorized_instrument():
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler

    inst = Assembler('inst', flavor=Flavor.MCSTAS)
    inst.parameter('double ex/"m"=0')
    inst.parameter('double phi/"degree"=0')

    inst.component('origin', 'Arm', at=(0, 0, 0))
    inst.component('source', 'Source_simple', at=[(0, 0, 0), 'origin'])
    inst.component('xpos', 'Arm', at=[('ex', 0, 0), 'source'])
    inst.component('zrot', 'Arm', at=[(0, 0, 0), 'xpos'], rotate=[(0, 0, 'phi'), 'xpos'])
    inst.component('aposrot', 'Arm', at=(1, 2, 3), rotate=(45, 55, 60))

    return inst.instrument


def test_motorized_instrument():
    import moreniius
    motorized = make_motorized_instrument()
    nx = moreniius.MorEniius.from_mccode(motorized, origin='origin', only_nx=False, absolute_depends_on=True)
    assert nx is not None
    #TODO add actual tests for the contents of, e.g., the dumped NeXus Structure
    ns = nx.to_nexus_structure()

    expected = {
        'entry': {
            'type': 'group',
            'children': 1,
            'next': 0,
            'attributes': [{'name': 'NX_class', 'dtype': 'string', 'values': 'NXentry'}]
        },
        'instrument': {
            'type': 'group',
            'children': 7,
            'next': 0,
            'attributes': [{'name': 'NX_class', 'dtype': 'string', 'values': 'NXinstrument'}]
        },
    }
    for name, has in expected.items():
        assert 'children' in ns
        assert len(ns['children']) >= has['next']
        ns = ns['children'][has['next']]
        assert all(x in ns for x in ('type', 'name', 'children', 'attributes'))
        assert ns['name'] == name
        assert ns['attributes'] == has['attributes']
        assert len(ns['children']) == has['children']

    xpos = ns['children'][3]
    zrot = ns['children'][4]
    aposrot = ns['children'][5]


    from json import dumps

    for cns in (xpos, zrot, aposrot):
        assert 'children' in cns
        assert 'transformations' in [c['name'] for c in cns['children'] if 'name' in c]
        t = [c for c in cns['children'] if 'name' in c and c['name'] == 'transformations'][0]
        assert 'children' in t
        t = t['children']
        for c in t:
            # Each child can _either_ be a dataset, with 'module' at its top level
            # Or a group, with 'name', etc. at its top level

            if 'module' in c:
                # this transformation is static, and a dataset
                assert 'dataset' == c['module']
                assert all(x in c for x in ('config', 'attributes'))
                assert all(x in c['config'] for x in ('name', 'values', 'type'))
                attrs = c['attributes']
                assert len(attrs) == 4
                assert all(all(x in a for x in ('name', 'values', 'dtype')) for a in attrs)
                assert all(a['name'] in ('vector', 'depends_on', 'transformation_type', 'units') for a in attrs)
            else:
                # this transformation is dynamic and a group
                assert all(x in c for x in ('name', 'type', 'children', 'attributes'))
                assert 'group' == c['type']
                attrs = c['attributes']
                assert len(attrs) == 1
                attr = attrs[0]
                assert all(x in attr for x in ('name', 'dtype', 'values'))
                assert 'NX_class' == attr['name']
                assert 'NXgroup' == attr['values']

                # The children should contain a link to the log ... is the order important?
                # Must the number of children always be the same?
                assert all('module' in cc for cc in c['children'])
                assert sum('link' == cc['module'] for cc in c['children']) <= 1
                for cc in c['children']:
                    if 'link' == cc['module']:
                        assert all(x in cc['config'] for x in ('name', 'source'))
                    else:
                        assert all(x in cc['config'] for x in ('name', 'values', 'type'))
                        assert cc['config']['name'] in ('vector', 'depends_on', 'transformation_type', 'units')

    # for name, has in expected.items():
    #     print({x : list(ns[x]) for x in ns})
    #     assert all(x in ns for x in has)
    #     assert len(ns['children']) == 1
    #     assert ns['name'] == name
    #     ns = ns['children'][0]

    #
    # instr = ns['entry']['instrument']
    # #
    # print(dumps(xpos, indent=2))
    # print(dumps(zrot, indent=1))
    # print(dumps(aposrot, indent=1))


"""This file holds the default translators for a subset of Component types, used by NXInstance.

They can serve as a template for adding more translators, or replacing defaults, as needed.

A translator is a function that takes a component NXInstance and returns a valid NeXus object
for the _component type_ of that instance. The NXInstance holds a reference to the NXInstr object,
which can be used to look up values of identifier parameters, and the instance mccode_antlr.instr.Instance
object, which holds the instance's parameters and component type, among other information.

For a translator

def translator(nxinstance: NXInstance):
    ...

registering it as an attribute of the NXInstance class will make it available as a translator for
all instances of that class. For example:

NXInstance.SomeComponentTypeName = translator

or

NXInstance.SomeComponentTypeName = staticmethod(translator)

or

setattr(NXInstance, 'SomeComponentTypeName', translator)

will all work, even if added to the class after instances have been created.
(But since the translation is done automatically by __post_init__, it will not be applied to existing instances.)

A very nice possible enhancement to this approach would be to register each translator as a property
of the NXInstance class; then it can be used 'transparently' without needing to explicitly call it.
"""
from zenlog import log
from mccode_antlr.common import Expr
from moreniius.utils import resolve_parameter_links


def slit_translator(nxinstance):
    """The Slit component _must_ define (xmin, xmax) _or_ xwidth, and similarly the y-named parameters"""
    from nexusformat.nexus import NXslit
    if nxinstance.obj.defines_parameter('xwidth'):
        x_gap = nxinstance.parameter('xwidth')
        x_zero = Expr.float(0)
    else:
        x_gap = nxinstance.parameter('xmax') - nxinstance.parameter('xmin')
        x_zero = nxinstance.parameter('xmax') + nxinstance.parameter('xmin')
    if nxinstance.obj.defines_parameter('ywidth'):
        y_gap = nxinstance.parameter('ywidth')
        y_zero = Expr.float(0)
    else:
        y_gap = nxinstance.parameter('ymax') - nxinstance.parameter('ymin')
        y_zero = nxinstance.parameter('ymax') + nxinstance.parameter('ymin')

    if isinstance(x_zero, Expr) or isinstance(y_zero, Expr):
        log.warn(f'{nxinstance.obj.name} has a non-constant x or y zero, which requires special handling for NeXus')
    elif abs(x_zero) or abs(y_zero):
        log.warn(f'{nxinstance.obj.name} should be translated by [{x_zero}, {y_zero}, 0] via eniius_data METADATA')
    params = resolve_parameter_links(dict(x_gap=x_gap, y_gap=y_gap))
    return nxinstance.make_nx(NXslit, **params)


def guide_translator(nxinstance):
    from nexusformat.nexus import NXguide
    from moreniius.nxoff import NXoff
    off_pars = {k: nxinstance.nx_parameter(k) for k in ('l', 'w1', 'h1', 'w2', 'h2')}
    for k in ('w', 'h'):
        off_pars[f'{k}2'] = off_pars[f'{k}1'] if off_pars[f'{k}2'] == 0 else off_pars[f'{k}2']
    guide_pars = {'m_value': nxinstance.parameter('m')}
    geometry = NXoff.from_wedge(**off_pars).to_nexus()
    return nxinstance.make_nx(NXguide, OFF_GEOMETRY=geometry, **resolve_parameter_links(guide_pars))


def collimator_linear_translator(nxinstance):
    from nexusformat.nexus import NXcollimator
    from moreniius.nxoff import NXoff
    pars = {k: nxinstance.nx_parameter(v) for k, v in (('l', 'length'), ('w1', 'xwidth'), ('h1', 'yheight'))}
    col_pars = dict(divergence_x=nxinstance.parameter('divergence'), divergence_y=nxinstance.parameter('divergenceV'))
    return nxinstance.make_nx(NXcollimator, OFF_GEOMETRY=NXoff.from_wedge(**pars).to_nexus(),
                              **resolve_parameter_links(col_pars))


def diskchopper_translator(nxinstance):
    from nexusformat.nexus import NXdisk_chopper, NXfield
    mpars = {k: nxinstance.parameter(k) for k in ('nslit', 'nu', 'radius', 'theta_0', 'phase', 'yheight')}
    pars = {'slits': mpars['nslit'],
            'rotation_speed': nxinstance.make_nx(NXfield, mpars['nu'], units='Hz'),
            'radius': nxinstance.make_nx(NXfield, mpars['radius'], units='m'),
            'slit_angle': nxinstance.make_nx(NXfield, mpars['theta_0'], units='degrees'),
            'phase': nxinstance.make_nx(NXfield, mpars['phase'], units='degrees'),
            'slit_height': nxinstance.make_nx(NXfield, mpars['yheight'] if mpars['yheight'] else mpars['radius'], units='m')}
    nslit, delta = mpars['nslit'], mpars['theta_0'] / 2.0
    slit_edges = [y * 360.0 / nslit + x for y in range(int(nslit)) for x in (-delta, delta)]
    nx_slit_edges = [nxinstance.expr2nx(se) for se in slit_edges]
    return nxinstance.make_nx(NXdisk_chopper, slit_edges=NXfield(nx_slit_edges, units='degrees'), **resolve_parameter_links(pars))


def elliptic_guide_gravity_translator(nxinstance):
    from nexusformat.nexus import NXguide
    from numpy import arange, sqrt
    from moreniius.nxoff import NXoff
    if not '"mid"' == nxinstance.obj.get_parameter('dimensionsAt'):
        log.warn('Only midpoint geometry supported by Elliptic_guide_gravity translator')
        log.info(f'The current guide has {nxinstance.obj.get_parameter("dimensionsAt")} specified')

    def ellipse_width(minor, distance, at):
        major = sqrt((distance / 2) ** 2 + minor ** 2)
        return 0 if abs(at) > major else minor * sqrt(1 - (at / major) ** 2)

    pars = dict(xw='xwidth', xi='linxw', xo='loutxw', yw='yheight', yi='linyh', yo='loutyh', l='l')
    p = {k: nxinstance.parameter(v) for k, v in pars.items()}
    n = 10
    rings = arange(n + 1) / n
    faces, vertices = [], []
    for x in rings:
        w = ellipse_width(p['xw'] / 2, p['xi'] + p['l'] + p['xo'], p['xi'] / 2 + (x - 0.5) * p['l'] - p['xo'] / 2)
        h = ellipse_width(p['yw'] / 2, p['yi'] + p['l'] + p['yo'], p['yi'] / 2 + (x - 0.5) * p['l'] - p['yo'] / 2)
        z = x * p['l']
        vertices.extend([[-w, -h, z], [-w, h, z], [w, h, z], [w, -h, z]])

    # These are only the guide faces (that is, the inner faces of the sides of the guide housing)
    # The entry and exit are not guide faces and therefore are NOT represented here!
    for i in range(n):
        j0, j1, j2, j3, j4, j5, j6, j7 = [4 * i + k for k in range(8)]
        faces.extend([[j0, j1, j5, j4], [j1, j2, j6, j5], [j2, j3, j7, j6], [j3, j0, j4, j7]])

    nx_vertices = [[nxinstance.expr2nx(expr) for expr in vector] for vector in vertices]
    nx_faces = [[nxinstance.expr2nx(expr) for expr in face] for face in faces]

    return NXguide(OFF_GEOMETRY=NXoff(nx_vertices, nx_faces).to_nexus())


def monitor_translator(nxinstance):
    from nexusformat.nexus import NXmonitor, NXdata
    from moreniius.nxoff import NXoff
    from moreniius.utils import NotNXdict
    from json import loads
    width = nxinstance.nx_parameter('xwidth')
    height = nxinstance.nx_parameter('yheight')
    geometry = NXoff.from_wedge(l=0.005, w1=width, h1=height)
    nx_monitor = NXmonitor(OFF_GEOMETRY=geometry.to_nexus())
    if len(nxinstance.obj.metadata):
        # look for mimetype 'application/json' and check if it is NeXus Structure data stream:
        for md in nxinstance.obj.metadata:
            if md.mimetype == 'application/json' and md.name == 'nexus_structure_stream_data':
                nx_monitor['data'] = NXdata(data=NotNXdict(loads(md.value)))

    return nx_monitor



from zenlog import log
from dataclasses import dataclass, field
from networkx import DiGraph
from typing import Union
from mccode_antlr.instr import Orient
from .instr import NXInstr

log.level('error')


@dataclass
class NXMcCode:
    nx_instr: NXInstr
    origin_name: Union[str, None] = None
    indexes: dict[str, int] = field(default_factory=dict)
    orientations: dict[str, Orient] = field(default_factory=dict)
    graph: Union[DiGraph, None] = None
    reversed_graph: Union[DiGraph, None] = None

    def __post_init__(self):
        from copy import deepcopy
        for index, instance in enumerate(self.nx_instr.instr.components):
            self.indexes[instance.name] = index
            self.orientations[instance.name] = deepcopy(instance.orientation)
        # Attempt to re-center all component dependent orientations on the sample
        found = (lambda x: self.origin_name == x.name) if self.origin_name else (lambda x: 'samples' == x.type.category)
        possible_origins = [instance for instance in self.nx_instr.instr.components if found(instance)]

        if not possible_origins:
            msg = '"sample" category components' if self.origin_name is None else f'component named {self.origin_name}'
            log.warn(f'No {msg} in instrument, using ABSOLUTE positions')
        elif self.origin_name is not None and len(possible_origins) > 1:
            log.error(f'{len(possible_origins)} components named {self.origin_name}; using the first')
        elif len(possible_origins) > 1:
            log.warn(f'More than one "sample" category component. Using {possible_origins[0].name} for origin name')
        if possible_origins:
            self.origin_name = possible_origins[0].name
            # find the position _and_ rotation of the origin
            origin = possible_origins[0].orientation
            # remove this from all components (re-centering on the origin)
            for name in self.orientations:
                self.orientations[name] = self.orientations[name] - origin

        if self.graph is None:
            self.graph = self.build_graph()
        if self.reversed_graph is None:
            self.reversed_graph = self.graph.reverse(copy=True)

    def transformations(self, name):
        from .orientation import NXOrient
        return NXOrient(self.nx_instr, self.orientations[name]).transformations(name)

    def inputs(self, name):
        """Return the other end of edges ending at the named node"""
        return list(self.reversed_graph[name])

    def outputs(self, name):
        """Return the other end of edges starting at the named node"""
        return list(self.graph[name])

    def component(self, name, only_nx=True):
        """Return a NeXus NXcomponent corresponding to the named McStas component instance"""
        from .instance import NXInstance
        instance = self.nx_instr.instr.components[self.indexes[name]]
        transformations = self.transformations(name)
        nxinst = NXInstance(self.nx_instr, instance, self.indexes[name], transformations, only_nx=only_nx)
        if transformations and nxinst.nx['transformations'] != transformations:
            # if the component modifed the transformations group, make sure we don't use our version again
            del self.orientations[name]
        if len(inputs := self.inputs(name)):
            nxinst.nx.attrs['inputs'] = inputs
        if len(outputs := self.outputs(name)):
            nxinst.nx.attrs['outputs'] = outputs
        return nxinst

    def instrument(self, only_nx=True):
        from nexusformat.nexus import NXinstrument
        nx = NXinstrument()  # this is a NeXus class
        nx['mcstas'] = self.nx_instr.to_nx()
        for name in self.indexes.keys():
            nx[name] = self.component(name, only_nx=only_nx).nx

        return nx

    def build_graph(self):
        # FIXME expand this to a full-description if/when McCode includes graph information
        graph = DiGraph()
        names = [x.name for x in self.nx_instr.instr.components]
        graph.add_nodes_from(names)
        # By default, any McCode instrument is a linear object:
        graph.add_edges_from([(names[i], names[i+1]) for i in range(len(names)-1)])
        return graph


from __future__ import annotations
from networkx import Graph
from mccode_antlr.instr import Instr


class MorEniius:
    def __init__(self, nx, only_nx: bool = False, absolute_depends_on: bool = False):
        self.nx = nx
        self.only_nx = only_nx
        self.absolute_depends_on = absolute_depends_on

    @classmethod
    def from_mccode(cls, instr: Instr,
                    origin: str | None = None,
                    only_nx: bool = False,
                    nxlog_root: str | None = None,
                    absolute_depends_on: bool = False,
                    graph: Graph | None = None,
                    ):
        from nexusformat.nexus import NXfield
        from .mccode import NXMcCode, NXInstr
        nxlog_root = nxlog_root or '/entry/parameters'
        nx_mccode = NXMcCode(NXInstr(instr, nxlog_root=nxlog_root), origin_name=origin, graph=graph)
        nxs_obj = nx_mccode.instrument(only_nx=only_nx)
        nxs_obj['name'] = NXfield(value=instr.name)
        return cls(nxs_obj, only_nx=only_nx, absolute_depends_on=absolute_depends_on)

    def to_nexus_structure(self):
        from .writer import Writer
        return Writer(self.nx).to_nexus_structure(only_nx=self.only_nx, absolute_depends_on=self.absolute_depends_on)

    def to_json(self, filename, indent=4):
        from .writer import Writer
        if not filename.endswith('.json'):
            filename += '.json'
        Writer(self.nx).to_json(filename, indent=indent, only_nx=self.only_nx, absolute_depends_on=self.absolute_depends_on)



import unittest

# TODO When github.com/g5t/mccode-to-kafka fully switched to using da00, these tests will fail.

class NexusStrctureTestCase(unittest.TestCase):
    def setUp(self):
        from json import dumps
        from mccode_antlr.loader import parse_mcstas_instr
        from mccode_to_kafka.writer import da00_variable_config, da00_dataarray_config
        t0 = {'name': 't', 'unit': 'usec', 'label': 'monitor', 'data': {'first': 0.5, 'last': 10.5, 'size': 11}}
        t1 = {'name': 't', 'unit': 'usec', 'label': 'monitor', 'data': {'first': 1.5, 'last': 11.5, 'size': 11}}
        m0 = da00_dataarray_config(topic='mon0', source='mccode-to-kafka', constants=[da00_variable_config(**t0)])
        m1 = da00_dataarray_config(topic='mon1', source='mccode-to-kafka', constants=[da00_variable_config(**t1)])

        instr = f"""DEFINE INSTRUMENT chopper_spectrometer(
        ch1speed, ch2speed, ch1phase, ch2phase
        )
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple() AT (0, 0, 0) RELATIVE origin
        COMPONENT mon0 = TOF_monitor(restore_neutron=1) AT (0, 0, 9) RELATIVE source
        METADATA "application/json" "nexus_structure_stream_data" %{{{dumps(m0)}%}}
        COMPONENT ch1 = DiskChopper(theta_0=170, radius=0.35, nu=ch1speed, phase=ch1phase) AT (0, 0, 10) RELATIVE source
        COMPONENT ch2 = DiskChopper(theta_0=170, radius=0.35, nu=ch2speed, phase=ch2phase) AT (0, 0, 0.1) RELATIVE ch1
        COMPONENT mon1 = TOF_monitor(restore_neutron=1) AT (0, 0, 0.1) RELATIVE ch2
        METADATA "application/json" "nexus_structure_stream_data" %{{{dumps(m1)}%}}
        COMPONENT sample = Arm() AT (0, 0, 80) RELATIVE ch2
        END
        """
        self.instr = parse_mcstas_instr(instr)
        self.structures = {'mon0': m0, 'mon1': m1}

    def test_moreniius(self):
        from moreniius import MorEniius
        from moreniius.utils import NotNXdict
        from nexusformat.nexus import NXdata, NXfield
        me = MorEniius.from_mccode(self.instr, origin='sample_stack', only_nx=False, absolute_depends_on=True)
        self.assertTrue(isinstance(me, MorEniius))
        for k in self.structures.keys():
            self.assertTrue(k in me.nx)
            self.assertTrue('data' in me.nx[k])
            a = me.nx[k]['data']
            self.assertTrue(hasattr(a, 'nxclass'))
            self.assertTrue(isinstance(a, NXdata))
            self.assertTrue(isinstance(a.data, NXfield))
            b = a.data.nxdata  # Why did a become an NXdata when it was an NXfield?
            self.assertTrue(isinstance(b, NotNXdict))
            c = b.to_json_dict()
            self.assertTrue(isinstance(c, dict))
            s = self.structures[k]
            self.assertEqual(c, s)

    def test_nexus_structure(self):
        from moreniius.nexus_structure import to_nexus_structure
        nx = to_nexus_structure(self.instr)
        self.assertTrue(isinstance(nx, dict))
        self.assertEqual(len(nx), 1)
        self.assertTrue('children' in nx)
        self.assertEqual(len(nx['children']), 1)
        nx = nx['children'][0]
        group_keys = ('name', 'type', 'children', 'attributes')
        for x in group_keys:
            self.assertTrue(x in nx)
        self.assertEqual(nx['name'], 'entry')
        self.assertEqual(nx['type'], 'group')
        self.assertEqual(len(nx['children']), 1)
        nx = nx['children'][0]
        for x in group_keys:
            self.assertTrue(x in nx)
        self.assertEqual(nx['name'], 'instrument')
        self.assertEqual(len(nx['children']), 9)
        nx = nx['children'][3]
        for x in group_keys:
            self.assertTrue(x in nx)
        self.assertEqual(nx['name'], 'mon0')
        self.assertEqual(len(nx['children']), 4)  # removed mcstas child
        nx = nx['children'][1] # this is now a NXdata group
        self.assertTrue('attributes' in nx)
        self.assertEqual(len(nx['attributes']), 1)
        self.assertEqual(nx['attributes'][0]['name'], 'NX_class')
        self.assertEqual(nx['attributes'][0]['values'], 'NXdata')
        nx = nx['children'][0]
        self.assertEqual(self.structures['mon0'], nx)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from zenlog import log
from dataclasses import dataclass
from mccode_antlr.instr import Orient, Parts, Part
from nexusformat.nexus import NXfield
from .instr import NXInstr

@dataclass
class NXPart:
    instr: NXInstr
    o: Part

    def expr2nx(self, expr):
        return self.instr.expr2nx(expr)

    def make_nx(self, nx_class, *args, **kwargs):
        return self.instr.make_nx(nx_class, *args, **kwargs)

    def make_translation(self, norm, vec, dep):
        # if `norm` is a link or NXlog, we should make a group not an NXfield
        return self.make_nx(NXfield, norm, vector=vec, depends_on=dep, transformation_type='translation', units='m')

    def translations(self, dep: str, name: str) -> list[tuple[str, NXfield]]:
        from mccode_antlr.instr import RotationPart
        from mccode_antlr.common import Expr, Value
        if isinstance(self.o, RotationPart):
            raise RuntimeError('Part is a rotation!')
        pos = self.o.position()
        if any(isinstance(c, (Expr, Value)) for c in (pos.x, pos.y, pos.z)):
            translations = []
            for n, c, v in (('x', pos.x, [1, 0, 0]), ('y', pos.y, [0, 1, 0]), ('z', pos.z, [0, 0, 1])):
                if c != Expr.parse('0'):
                    next_name = f'{name}_{n}'
                    translations.append((next_name, self.make_translation(c, v, dep)))
                    dep = next_name
            return translations
        # vector is all constants, hopefully
        norm = pos.length()
        vec = pos if norm.is_zero else pos/norm
        return [(name, self.make_translation(norm, vec, dep))]

    def rotation(self, dep: str) -> NXfield:
        from mccode_antlr.instr import TranslationPart
        if isinstance(self.o, TranslationPart):
            raise RuntimeError('Part is a translation')
        try:
            axis, angle, angle_unit = self.o.rotation_axis_angle
        except RuntimeError as error:
            log.error(f'Failed to get rotation axis and angle: {error}')
            print(repr(self.o))
            raise NotImplementedError()

        # handle the case where angle is not a constant?
        return self.make_nx(NXfield, angle, vector=axis, depends_on=dep, transformation_type='rotation', units=angle_unit)

    def transformations(self, name: str, dep: str | None = None) -> list[tuple[str, NXfield]]:
        if self.o.is_translation and self.o.is_rotation:
            ops = self.translations(dep, name)
            rot = self.rotation(ops[-1][0])
            return [*ops, (f'{name}_r', rot)]
        elif self.o.is_translation:
            return self.translations(dep, name)
        elif self.o.is_rotation:
            return [(name, self.rotation(dep))]
        else:
            return []


@dataclass
class NXParts:
    instr: NXInstr
    position: Parts
    rotation: Parts

    def transformations(self, name: str) -> list[tuple[str, NXfield]]:
        nxt = []
        dep = '.'
        for index, o in enumerate(self.position.stack()):
            nxt.extend(NXPart(self.instr, o).transformations(f'{name}_t{index}', dep))
            dep = nxt[-1][0] if len(nxt) and len(nxt[-1]) else '.'
        for index, o in enumerate(self.rotation.stack()):
            nxt.extend(NXPart(self.instr, o).transformations(f'{name}_r{index}', dep))
            dep = nxt[-1][0] if len(nxt) and len(nxt[-1]) else '.'
        return nxt


@dataclass
class NXOrient:
    instr: NXInstr
    do: Orient

    def transformations(self, name: str) -> dict[str, NXfield]:
        # collapse all possible chained orientation information
        # But keep the rotations and translations separate
        pos, rot = self.do.position_parts(), self.do.rotation_parts()
        # make an ordered list of the requisite NXfield entries
        nxt = NXParts(self.instr, pos, rot).transformations(name)
        return {k: v for k, v in nxt}
